//The vertex shader that performs coordinate transformations, and projects the 3D to 2D coordinates.
var vertexShaderSource = `#version 300 es
	precision mediump float;

	//##########################################
	//#          Per-Vertex Inputs             #
	//##########################################

	//The vertex's 3D position in model space.
	in vec3 vertPos;
	
	//The 2D texture coordinates for this vertex (use for displacement mapping)
	in vec2 vertUV;
	
	//The normal vector at this vertex in model space.
	in vec3 vertNormal;
	
	//##########################################
	//#                Outputs                 #
	//##########################################

	//The colour for this vertex. By default, colours will be
	//smoothly interpolated. Add the word "flat" in front of "out" here and
	//in the fragment shader to do flat shading.
	flat out vec3 outCol;
	
	//The homogeneous coordinates of the vertex after projection.
	out vec4 outPosition;
	
	//##########################################
	//#     Displacement Mapping Parameters    #
	//##########################################
	
	//A weighting factor to multiply the displacement from the texture map
	const float displacementScale = 0.1;

	//A sampler to select pixels from the 2D texture defining the displacement map
	uniform sampler2D displacementMapTexture;
	
	//##########################################
	//#      Transformation Matrices           #
	//##########################################

	//Uniform variables are the same across all invocations of a shader
	//so each vertex will get the same set of transformation matrices:
	uniform mat4 modelMatrix;
	uniform mat4 viewMatrix;
	uniform mat4 projectionMatrix;

	//##########################################
	//#            Light Parameters            #
	//##########################################
	
	//Homogeneous coordinate of eye/camera (in world coordinates).
	const vec4 eyePos = vec4(0.0);
	
	//Homogeneous coordinate of light source  (in world coordinates).
	const vec4 lightPos = vec4(1.0, 0.0, 0.0, 1.0);

	//The single point light's intensity (split into rgb channels to give it a configurable colour).
	const vec3 lightIntensity = vec3(0.8);

	//The rgb colour of the ambient light in the scene.
	const vec3 lightAmbient = vec3(0.3);
	
	//##########################################
	//#           Material Parameters          #
	//##########################################

	//The diffuse colour of the model (how much diffuse light it reflects in rgb channels).
	const vec3 materialDiffuse = vec3(0.4, 0.9, 0.4);

	//The model's specular reflection intensity for each rgb colour channel.
	const vec3 materialSpecular = vec3(1.0, 1.0, 1.0);

	//The intensity of the ambient light reflected by the model for each rgb colour channel.
	const vec3 materialAmbient = vec3(0.4, 0.9, 0.4);

	//The exponent in the phong specularity calculation.
	//Smoother, more mirror-like surfaces should have higher values, 
	//which will reduce the size of the specular highlight.
	const float shininess = 10.0;

	
	//##########################################
	//#                 Main                   #
	//##########################################

	//The transformed vertex coordinates are passed on to the next stage of the graphics pipeline.
	void main() {
		float d = displacementScale * texture(displacementMapTexture, vertUV).x;
		vec3 displacement = d * vertNormal;

		vec4 pos = vec4(vertPos, 1.0);

		//Get the vertex's world-space coordinates (so that it is in the same coordinates that the eye and light are given in).
		vec4 vGlobal = modelMatrix * pos;

		//Calculate the vector from the vertex to the viewer's eye (in world coordinates).
		vec4 eyeVec = eyePos - vGlobal;

		//Calculate the vector from the vertex to the light source (in world coordinates).
		vec4 lightVec = lightPos - vGlobal;
		
		//Normalize all the input vectors to make sure they are all of
		//unit length (the linear interpolation does not preserve vector length).
		vec3 n = normalize((transpose(inverse(mat3(modelMatrix))) * vertNormal));
		vec3 l = normalize(lightVec.xyz);
		vec3 v = normalize(eyeVec.xyz);
		
		//Calculate n dot l in one place so it can be re-used without re-calculating it.
		float nDotL = dot(n,l);
		
		//Get the reflection vector "r"  of the light in direction "l" reflecting around the normal vector "n".
		vec3 r = normalize(2.0 * n * nDotL - l);

		//Calculate the diffuse, specular, and ambient colour values reflected from the surface
		//using the phong lighting equations.
		//Using max(0.0, ...) ensures that the dot products are non-negative 
		//(so light going in the opposite direction does not affect the calculation).
		//Vec3s are used to calculate the rgb components in a single step, rather than 3 separate calculations.
		
		vec3 diffuse = materialDiffuse * max(0.0, nDotL);
		vec3 specular = materialSpecular * pow(max(0.0, dot(v, r)), shininess);
		vec3 ambient = lightAmbient * materialAmbient;
		
		//Combine the ambient, diffuse and specular terms, and clamp them to below 1.0 (the max colour intensity).
		vec3 colour = min(ambient + lightIntensity * (diffuse + specular), 1.0);

		outCol = colour;

		gl_Position = projectionMatrix * viewMatrix * modelMatrix * pos; 
	}
`;

//This simple fragment shader makes every pixel it receives a single colour.
var fragmentShaderSource = `#version 300 es
	precision mediump float;

	flat in vec3 outCol;
	
	out vec4 fragColor;

	//Output the colour as an RGBA vector with full alpha
	void main() {
	  fragColor = vec4(outCol, 1.0);
	}
`;


//Compile and link the specified vertex and fragment shaders,
//throwing exceptions if either stage goes wrong.
function createProgram(gl, shaderSpecs) {
	var program = gl.createProgram();
	
	//Go through all shaders in the list
	for(var i = 0; i < shaderSpecs.length; i++){
		var spec = shaderSpecs[i];
		var shader = gl.createShader(spec.type);
		
		//Load the source code from the script tags
		gl.shaderSource(shader, spec.shaderString);
		
		//Compile the shader and check for errors
		gl.compileShader(shader);
		if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
			throw gl.getShaderInfoLog(shader);
		}
		
		//Add the shader to the program, and clean up resources
		gl.attachShader(program, shader);
		gl.deleteShader(shader);
	}
	
	//Link the vertex and fragment shader stages together in the program.
	//If the stages do not have matching input and output variables, an error is thrown.
	gl.linkProgram(program);
	if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
		throw gl.getProgramInfoLog(program);
	}
	return program;
}


